#include "aicontrol.h"
#include "catalogue.h"
#include "layout.h"
#include "train.h"

using namespace std;
using namespace Msp;

namespace R2C2 {

AIControl::AIControl(Train &t):
	TrainAI(t),
	target_speed(0),
	reverse(false),
	pending_reverse(false),
	state(NORMAL),
	need_update(false)
{
	train.signal_ai_event.connect(sigc::mem_fun(this, &AIControl::event));
}

void AIControl::set_target_speed(float s)
{
	target_speed = s;
	need_update = true;
	signal_event.emit(Message("target-speed-changed", target_speed));
}

void AIControl::set_reverse(bool r)
{
	pending_reverse = r;
	if(train.get_controller().get_speed())
		set_target_speed(0);
	else
	{
		reverse = r;
		train.set_control("reverse", reverse);
		signal_event.emit(Message("reverse-changed", reverse));
	}
}

void AIControl::message(const Message &msg)
{
	if(msg.type=="set-target-speed")
		set_target_speed(msg.value.value<float>());
	else if(msg.type=="set-reverse")
		set_reverse(msg.value.value<bool>());
	else if(msg.type=="toggle-reverse")
		set_reverse(!reverse);
}

void AIControl::tick(const Time::TimeDelta &)
{
	float scale = train.get_layout().get_catalogue().get_scale();
	float rsv_dist = train.get_reserved_distance();
	float brake_dist = train.get_controller().get_braking_distance();
	float approach_margin = 50*scale;
	float approach_speed = 5*scale;
	float margin = 1*scale;

	State old_state = state;

	if(state==FOLLOW && !train.get_preceding_train())
		state = NORMAL;

	if(rsv_dist<brake_dist+margin)
		state = BLOCKED;
	else if(rsv_dist>brake_dist+margin*2 && rsv_dist<brake_dist*1.3+approach_margin)
		state = APPROACH;
	else if(rsv_dist>brake_dist*1.3+approach_margin*2)
		state = NORMAL;

	if(state==NORMAL && train.get_preceding_train())
		state = FOLLOW;

	if(state!=old_state || state==FOLLOW || need_update)
	{
		float speed_limit = -1;
		if(state==BLOCKED)
			speed_limit = 0;
		else if(state==APPROACH)
			speed_limit = approach_speed;
		else if(state==FOLLOW && train.get_preceding_train()->get_block_allocator().is_active())
			speed_limit = train.get_preceding_train()->get_speed();

		if(speed_limit>=0 && target_speed>speed_limit)
			train.set_control("speed", speed_limit);
		else
			train.set_control("speed", target_speed);

		need_update = false;
	}

	if(pending_reverse!=reverse && !train.get_controller().get_speed())
	{
		reverse = pending_reverse;
		train.set_control("reverse", reverse);
	}
}

bool AIControl::has_intent_to_move() const
{
	return target_speed;
}

void AIControl::event(TrainAI &, const Message &ev)
{
	if(ev.type=="arrived")
		set_target_speed(0);
}

} // namespace R2C2
