#include <msp/core/maputils.h>
#include <msp/time/utils.h>
#include "catalogue.h"
#include "layout.h"
#include "route.h"
#include "train.h"
#include "trainroutemetric.h"
#include "trainrouteplanner.h"
#include "trainrouter.h"
#include "vehicle.h"

using namespace std;
using namespace Msp;

namespace R2C2 {

TrainRoutePlanner::TrainRoutePlanner(Layout &layout):
	goal(0),
	timeout(10*Time::sec),
	result(PENDING),
	thread(0)
{
	const map<unsigned, Train *> &trains = layout.get_trains();
	for(map<unsigned, Train *>::const_iterator i=trains.begin(); i!=trains.end(); ++i)
	{
		TrainRoutingInfo info(*i->second);
		if(!info.waypoints.empty())
			routed_trains.push_back(info);
	}
}

TrainRoutePlanner::~TrainRoutePlanner()
{
	if(thread)
	{
		thread->join();
		delete thread;
	}
}

void TrainRoutePlanner::set_timeout(const Time::TimeDelta &t)
{
	timeout = t;
}

TrainRoutePlanner::Result TrainRoutePlanner::plan()
{
	prepare_plan();
	create_plan();
	if(result==PENDING)
		finalize_plan();

	return result;
}

void TrainRoutePlanner::plan_async()
{
	if(thread)
		throw logic_error("already planning");

	prepare_plan();
	thread = new PlanningThread(*this);
}

TrainRoutePlanner::Result TrainRoutePlanner::check()
{
	if(result==PENDING && goal)
	{
		if(thread)
		{
			thread->join();
			delete thread;
			thread = 0;
		}
		finalize_plan();
	}

	return result;
}

const list<Route *> &TrainRoutePlanner::get_routes_for(const Train &train) const
{
	return get_train_info(train).routes;
}

const list<TrainRouter::SequencePoint> &TrainRoutePlanner::get_sequence_for(const Train &train) const
{
	return get_train_info(train).sequence;
}

const TrainRoutePlanner::TrainRoutingInfo &TrainRoutePlanner::get_train_info(const Train &train) const
{
	for(vector<TrainRoutingInfo>::const_iterator i=routed_trains.begin(); i!=routed_trains.end(); ++i)
		if(i->train==&train)
			return *i;

	throw key_error(train.get_name());
}

const TrainRoutePlanner::RoutingStep &TrainRoutePlanner::get_step()
{
	steps.splice(steps.end(), queue, queue.begin());
	return steps.back();
}

void TrainRoutePlanner::prepare_plan()
{
	steps.clear();
	queue.clear();
	goal = 0;
	result = PENDING;

	queue.push_back(RoutingStep());
	RoutingStep &start = queue.back();
	for(vector<TrainRoutingInfo>::iterator i=routed_trains.begin(); i!=routed_trains.end(); ++i)
		start.trains.push_back(TrainRoutingState(*i));
	start.update_estimate();
}

void TrainRoutePlanner::create_plan()
{
	Time::TimeStamp timeout_stamp = Time::now()+timeout;
	unsigned count = 0;
	while(!queue.empty())
	{
		const RoutingStep &step = get_step();
		if(step.is_goal())
		{
			goal = &step;
			return;
		}

		add_steps(step);

		if(++count>=1000)
		{
			if(Time::now()>timeout_stamp)
				break;
			count = 0;
		}
	}

	result = FAILED;
}

void TrainRoutePlanner::add_steps(const RoutingStep &step)
{
	list<RoutingStep> new_steps;
	step.create_successors(new_steps);
	new_steps.sort();
	queue.merge(new_steps);
}

void TrainRoutePlanner::finalize_plan()
{
	for(vector<TrainRoutingInfo>::iterator i=routed_trains.begin(); i!=routed_trains.end(); ++i)
	{
		i->routes.clear();
		i->sequence.clear();
		for(unsigned j=0; j<2; ++j)
			i->track_history[j] = 0;
	}

	map<Track *, TrainRouter::SequencePoint *> sequenced_tracks;
	unsigned sequence = steps.size();
	for(const RoutingStep *i=goal; i; i=i->prev)
		for(vector<TrainRoutingState>::const_iterator j=i->trains.begin(); j!=i->trains.end(); ++j)
		{
			Track **history = j->info->track_history;
			// Don't process the same track again.
			if(j->track.track()==history[0])
				continue;

			Route *route = 0;
			bool start_new_route = true;
			if(!j->info->routes.empty())
			{
				/* If we already have a route and this track or any linked track is
				in it, start a new one to avoid loops. */
				route = j->info->routes.front();
				start_new_route = route->has_track(*j->track);
				if(!start_new_route)
				{
					unsigned nls = j->track->get_n_link_slots();
					for(unsigned k=0; (!start_new_route && k<nls); ++k)
					{
						Track *link = j->track->get_link(k);
						start_new_route = (link && link!=history[0] && route->has_track(*link));
					}
				}
			}

			if(start_new_route)
			{
				route = new Route(j->info->train->get_layout());
				route->set_name("Router");
				route->set_temporary(true);
				/* Have the routes overlap by two tracks to ensure that turnout
				paths can be deduced. */
				for(unsigned k=0; (k<2 && history[k]); ++k)
					route->add_track(*history[k]);
				j->info->routes.push_front(route);
			}

			route->add_track(*j->track.track());
			history[1] = history[0];
			history[0] = j->track.track();

			bool waitable = j->track.endpoint().paths!=j->track->get_type().get_paths();
			map<Track *, TrainRouter::SequencePoint *>::iterator k = sequenced_tracks.find(j->track.track());
			if(k!=sequenced_tracks.end())
			{
				// Add a sequence point if another train uses this track afterwards.
				if(!k->second->preceding_train)
				{
					k->second->preceding_train = j->info->train;
					k->second->sequence_in = sequence;
				}
				j->info->sequence.push_front(TrainRouter::SequencePoint(j->track->get_block(), sequence));
				if(waitable)
					k->second = &j->info->sequence.front();
				--sequence;
			}
			else if(waitable)
			{
				/* Create a sequence point if it's possible to wait and let another
				train past. */
				j->info->sequence.push_front(TrainRouter::SequencePoint(j->track->get_block(), sequence));
				sequenced_tracks[j->track.track()] = &j->info->sequence.front();
				--sequence;
			}
		}

	result = COMPLETE;
}


TrainRoutePlanner::TrainRoutingInfo::TrainRoutingInfo(Train &t):
	train(&t),
	speed(train->get_maximum_speed()),
	first_noncritical(train->get_last_critical_block().next().block()),
	router(train->get_ai_of_type<TrainRouter>()),
	has_duration(false)
{
	if(unsigned n_wps = router->get_n_waypoints())
	{
		waypoints.reserve(n_wps),
		metrics.reserve(n_wps);
		for(unsigned i=0; i<n_wps; ++i)
		{
			waypoints.push_back(router->get_waypoint(i));
			metrics.push_back(&router->get_metric(i));
		}
		has_duration = router->get_trip_duration();
	}

	// If no maximum speed is specified, use a sensible default
	if(!speed)
		speed = 20*train->get_layout().get_catalogue().get_scale();
}


TrainRoutePlanner::OccupiedTrack::OccupiedTrack(Track &t, unsigned p, OccupiedTrack *n):
	track(&t),
	path_length(track->get_type().get_path_length(p)),
	next(n),
	n_tracks(next ? next->n_tracks+1 : 1),
	refcount(1)
{
	if(next)
		++next->refcount;
}

TrainRoutePlanner::OccupiedTrack::OccupiedTrack(const OccupiedTrack &other):
	track(other.track),
	path_length(other.path_length),
	next(other.next),
	n_tracks(other.n_tracks),
	refcount(1)
{
	if(next)
		++next->refcount;
}

TrainRoutePlanner::OccupiedTrack::~OccupiedTrack()
{
	if(next && !--next->refcount)
		delete next;
}


TrainRoutePlanner::TrainRoutingState::TrainRoutingState(TrainRoutingInfo &inf):
	info(&inf),
	critical(true),
	occupied_tracks(0),
	state(MOVING),
	delay(info->router->get_departure_delay()),
	duration(info->router->get_trip_duration()),
	waypoint(0),
	blocked_by(-1)
{
	const Vehicle *veh = &info->train->get_vehicle(0);
	// TODO margins
	TrackOffsetIter track_and_offs = veh->get_placement().get_position(VehiclePlacement::FRONT_BUFFER);
	track = track_and_offs.track_iter();
	offset = track_and_offs.offset();
	path = track->get_active_path();

	while(Vehicle *next = veh->get_link(1))
		veh = next;
	track_and_offs = veh->get_placement().get_position(VehiclePlacement::BACK_BUFFER);
	back_offset = track_and_offs.offset();

	TrackIter iter = track_and_offs.track_iter();
	while(1)
	{
		occupied_tracks = new OccupiedTrack(*iter, iter->get_active_path(), occupied_tracks);
		if(iter.track()==track.track())
			break;
		iter = iter.next();
	}

	update_estimate();
}

TrainRoutePlanner::TrainRoutingState::TrainRoutingState(const TrainRoutingState &other):
	info(other.info),
	track(other.track),
	path(other.path),
	critical(other.critical),
	occupied_tracks(other.occupied_tracks),
	offset(other.offset),
	back_offset(other.back_offset),
	state(other.state),
	delay(other.delay),
	duration(other.duration),
	waypoint(other.waypoint),
	distance_traveled(other.distance_traveled),
	remaining_estimate(other.remaining_estimate),
	wait_time(other.wait_time),
	blocked_by(other.blocked_by)
{
	++occupied_tracks->refcount;
}

TrainRoutePlanner::TrainRoutingState::~TrainRoutingState()
{
	if(occupied_tracks && !--occupied_tracks->refcount)
		delete occupied_tracks;
}

Time::TimeDelta TrainRoutePlanner::TrainRoutingState::get_time_to_next_track() const
{
	return ((track->get_type().get_path_length(path)-offset)/info->speed)*Time::sec+delay;
}

bool TrainRoutePlanner::TrainRoutingState::is_occupying(Track &trk) const
{
	if(state==ARRIVED && !duration && info->has_duration)
		return false;

	OccupiedTrack *occ = occupied_tracks;
	for(unsigned n=occ->n_tracks; n>0; --n, occ=occ->next)
		if(occ->track==&trk)
			return true;
	return false;
}

bool TrainRoutePlanner::TrainRoutingState::check_arrival()
{
	TrackIter next_track = track.next(path);

	// Check if we're about the exit the current waypoint's tracks.
	const TrainRouter::Waypoint &wp = info->waypoints[waypoint];
	if(wp.chain->has_track(*track) && !wp.chain->has_track(*next_track))
		if(wp.direction==TrackChain::UNSPECIFIED || track==wp.chain->iter_for(*track, wp.direction))
		{
			if(waypoint+1<info->waypoints.size())
				++waypoint;
			else
			{
				state = ARRIVED;
				return true;
			}
		}

	// If we're entering the first non-critical block, clear the critical flag.
	if(info->first_noncritical->has_track(*next_track))
		critical = false;

	return false;
}

void TrainRoutePlanner::TrainRoutingState::advance(float distance)
{
	offset += distance;
	back_offset += distance;

	// See if the tail end of the train has passed any sensors.
	unsigned count_to_free = 0;
	unsigned last_sensor_addr = 0;
	float distance_after_sensor = 0;
	OccupiedTrack *occ = occupied_tracks;
	for(unsigned n=occupied_tracks->n_tracks; n>0; --n)
	{
		if(unsigned saddr = occ->track->get_sensor_address())
		{
			if(saddr!=last_sensor_addr)
			{
				count_to_free = 0;
				distance_after_sensor = 0;
			}
			last_sensor_addr = saddr;
		}

		++count_to_free;
		distance_after_sensor += occ->path_length;

		occ = occ->next;
	}

	// Free the last passed sensor and any tracks behind it.
	if(count_to_free && back_offset>distance_after_sensor)
	{
		back_offset -= distance_after_sensor;
		if(occupied_tracks->refcount>1)
		{
			--occupied_tracks->refcount;
			occupied_tracks = new OccupiedTrack(*occupied_tracks);
		}
		occupied_tracks->n_tracks -= count_to_free;
	}

	distance_traveled += distance;
	remaining_estimate -= distance;
}

void TrainRoutePlanner::TrainRoutingState::advance(const Time::TimeDelta &dt)
{
	if(delay>=dt)
	{
		delay -= dt;
		return;
	}

	float secs = dt/Time::sec;
	// There may be negative delay remaining after previous step.
	if(delay)
	{
		secs -= delay/Time::sec;
		delay = Time::zero;
	}

	if(duration)
		duration = max(duration-secs*Time::sec, Time::zero);

	if(state==MOVING)
		advance(info->speed*secs);
	else if(state!=ARRIVED)
		wait_time += secs*Time::sec;
}

void TrainRoutePlanner::TrainRoutingState::advance_track(unsigned next_path)
{
	float distance = occupied_tracks->path_length-offset;
	track = track.next(path);
	path = next_path;
	occupied_tracks = new OccupiedTrack(*track, path, occupied_tracks);
	advance(distance);
	offset = 0;
}

void TrainRoutePlanner::TrainRoutingState::update_estimate()
{
	TrackIter iter = track.reverse(path);
	remaining_estimate = info->metrics[waypoint]->get_distance_from(*iter.track(), iter.entry());
	if(remaining_estimate>=0)
		remaining_estimate += track->get_type().get_path_length(path)-offset;
}

bool TrainRoutePlanner::TrainRoutingState::is_viable() const
{
	if(remaining_estimate<0)
		return false;
	if(critical && state==BLOCKED)
		return false;
	return true;
}


TrainRoutePlanner::RoutingStep::RoutingStep():
	prev(0)
{ }

TrainRoutePlanner::RoutingStep::RoutingStep(const RoutingStep *p):
	time(p->time),
	penalty(p->penalty),
	cost_estimate(p->cost_estimate),
	trains(p->trains),
	prev(p)
{ }

void TrainRoutePlanner::RoutingStep::create_successors(list<RoutingStep> &new_steps) const
{
	RoutingStep next(this);
	if(next.update_states() && next.check_deadlocks())
		return;

	int train_index = find_next_train();
	if(train_index<0)
		return;

	TrainRoutingState &train = next.trains[train_index];

	Time::TimeDelta dt = train.get_time_to_next_track();
	next.advance(dt);

	/* Check arrival after the train has advanced to the end of its current track
	so travel time and occupied tracks will be correct. */
	if(train.check_arrival())
	{
		new_steps.push_back(next);
		return;
	}

	train.advance_track(0);

	const TrackType::Endpoint &entry_ep = train.track.endpoint();
	if(train.critical)
	{
		/* Only create a successor step matching the currently set path for a
		critical track. */
		unsigned critical_path = train.track->get_type().coerce_path(train.track.entry(), train.track->get_active_path());
		create_successor(next, train_index, critical_path, new_steps);
	}
	else
	{
		// Create successor steps for all possible paths through the new track.
		for(unsigned i=0; entry_ep.paths>>i; ++i)
			if(entry_ep.has_path(i))
				create_successor(next, train_index, i, new_steps);
	}

	new_steps.sort();
	for(list<RoutingStep>::iterator i=new_steps.begin(); ++i!=new_steps.end(); )
	{
		i->penalty += 5*Time::sec;
		i->update_estimate();
	}

	if(entry_ep.paths!=train.track->get_type().get_paths() && !train.critical)
	{
		/* Create a waiting state before the track if there's at least one path
		that doesn't pass through the entry endpoint. */
		RoutingStep wait(this);
		wait.advance(dt);
		wait.trains[train_index].state = WAITING;
		wait.penalty += 15*Time::sec;
		wait.update_estimate();
		if(wait.is_viable())
			new_steps.push_back(wait);
	}
}

void TrainRoutePlanner::RoutingStep::create_successor(RoutingStep &next, unsigned train_index, unsigned path, list<RoutingStep> &new_steps)
{
	TrainRoutingState &train = next.trains[train_index];

	train.path = path;
	train.update_estimate();
	next.update_estimate();
	if(next.is_viable())
		new_steps.push_back(next);
}

bool TrainRoutePlanner::RoutingStep::update_states()
{
	bool changes = false;
	for(vector<TrainRoutingState>::iterator i=trains.begin(); i!=trains.end(); ++i)
	{
		if(i->state==ARRIVED)
			continue;

		TrainState old_state = i->state;

		TrackIter next_track = i->track.next(i->path);
		if(next_track)
		{
			i->blocked_by = get_occupant(*next_track);
			if(i->blocked_by>=0)
			{
				/* If the train is still traversing its last critical track, the
				flag needs to be cleared here to pass viability test. */
				if(i->info->first_noncritical->has_track(*next_track))
					i->critical = false;

				/* Trains in the WAITING state will also transition to BLOCKED and
				then to MOVING when the other train has passed. */
				i->state = BLOCKED;
			}
			else if(i->state==BLOCKED)
				i->state = MOVING;
		}
		else
			i->state = BLOCKED;

		if(i->state!=old_state)
			changes = true;
	}

	return changes;
}

bool TrainRoutePlanner::RoutingStep::check_deadlocks() const
{
	for(vector<TrainRoutingState>::const_iterator i=trains.begin(); i!=trains.end(); ++i)
	{
		if(i->state!=BLOCKED)
			continue;

		// A train blocked by end of track is always considered a deadlock.
		if(i->blocked_by<0)
			return true;

		/* Use the tortoise and hare algorithm to check if trains are blocked
		cyclically (A blocks B, which blocks ..., which blocks A). */
		int slow = i->blocked_by;
		int fast = trains[slow].blocked_by;
		while(fast>=0 && trains[fast].blocked_by>=0)
		{
			if(fast==slow)
				return true;

			slow = trains[slow].blocked_by;
			fast = trains[trains[fast].blocked_by].blocked_by;
		}
	}

	return false;
}

int TrainRoutePlanner::RoutingStep::get_occupant(Track &track) const
{
	for(unsigned i=0; i<trains.size(); ++i)
		if(trains[i].is_occupying(track))
			return i;

	return -1;
}

int TrainRoutePlanner::RoutingStep::find_next_train() const
{
	Time::TimeDelta min_dt;
	int next_train = -1;
	for(unsigned i=0; i<trains.size(); ++i)
		if(trains[i].state==MOVING)
		{
			Time::TimeDelta dt = trains[i].get_time_to_next_track();
			if(dt<min_dt || next_train<0)
			{
				min_dt = dt;
				next_train = i;
			}
		}

	return next_train;
}

void TrainRoutePlanner::RoutingStep::advance(const Time::TimeDelta &dt)
{
	time += dt;
	for(vector<TrainRoutingState>::iterator i=trains.begin(); i!=trains.end(); ++i)
		i->advance(dt);
}

void TrainRoutePlanner::RoutingStep::update_estimate()
{
	cost_estimate = penalty;
	for(vector<TrainRoutingState>::const_iterator i=trains.begin(); i!=trains.end(); ++i)
		if(i->remaining_estimate>=0)
			cost_estimate += i->wait_time+((i->distance_traveled+i->remaining_estimate)/i->info->speed)*Time::sec;
}

bool TrainRoutePlanner::RoutingStep::is_viable() const
{
	for(vector<TrainRoutingState>::const_iterator i=trains.begin(); i!=trains.end(); ++i)
		if(!i->is_viable())
			return false;

	for(vector<TrainRoutingState>::const_iterator i=trains.begin(); i!=trains.end(); ++i)
		if(i->state==MOVING)
			return true;

	return false;
}

bool TrainRoutePlanner::RoutingStep::is_goal() const
{
	for(vector<TrainRoutingState>::const_iterator i=trains.begin(); i!=trains.end(); ++i)
		if(i->state!=ARRIVED)
			return false;
	return true;
}

bool TrainRoutePlanner::RoutingStep::operator<(const RoutingStep &other) const
{
	return cost_estimate<other.cost_estimate;
}


TrainRoutePlanner::PlanningThread::PlanningThread(TrainRoutePlanner &p):
	planner(p)
{
	launch();
}

void TrainRoutePlanner::PlanningThread::main()
{
	planner.create_plan();
}

} // namespace R2C2
