#ifndef LIBR2C2_TRAINROUTER_H_
#define LIBR2C2_TRAINROUTER_H_

#include <list>
#include <msp/datafile/objectloader.h>
#include "trackchain.h"
#include "trainai.h"

namespace R2C2 {

class Block;
class Layout;
class Track;
class TrainRouteMetric;
class TrainRoutePlanner;

class TrainRouter: public TrainAI
{
public:
	class Loader: public Msp::DataFile::ObjectLoader<TrainRouter>
	{
	public:
		Loader(TrainRouter &);
	private:
		void route(const std::string &);
	};

	struct Waypoint
	{
		const TrackChain *chain;
		TrackChain::Direction direction;

		Waypoint(const TrackChain &, TrackChain::Direction = TrackChain::UNSPECIFIED);
	};

	struct SequencePoint
	{
		Block *block;
		Train *preceding_train;
		unsigned sequence_in;
		unsigned sequence_out;

		SequencePoint(Block &, unsigned);

		bool is_cleared() const;
	};

private:
	typedef std::list<const Route *> RouteList;

	enum TravelState
	{
		ON_THE_WAY,
		WAITING_FOR_SEQUENCE,
		SEQUENCE_CHECK_PENDING,
		RESERVED_TO_END,
		ADVANCED_TO_END,
		ARRIVED
	};

public:
	sigc::signal<void, const Route *> signal_route_changed;
	sigc::signal<void, const TrackChain *> signal_arrived;
	sigc::signal<void, const TrackChain *> signal_waypoint_reached;

private:
	int priority;
	RouteList routes;
	RouteList::iterator reserving_route;
	TravelState state;
	std::vector<Waypoint> waypoints;
	bool waypoints_changed;
	std::vector<TrainRouteMetric *> metrics;
	bool metrics_stale;
	std::list<SequencePoint> sequence_points;
	unsigned current_sequence;
	Msp::Time::TimeDelta delay;
	Msp::Time::TimeDelta duration;
	Msp::RefPtr<TrainRoutePlanner> planner;

public:
	TrainRouter(Train &);
	~TrainRouter();

	void set_priority(int);
	int get_priority() const { return priority; }

	bool set_route(const Route *);
	const Route *get_route() const;
	unsigned get_current_sequence() const { return current_sequence; }
private:
	void use_planned_route();
	void route_changed();

public:
	void set_destination(const TrackChain &);
	const TrackChain *get_destination() const { return waypoints.empty() ? 0 : waypoints.back().chain; }
	void add_waypoint(const TrackChain &, TrackChain::Direction = TrackChain::UNSPECIFIED);
	unsigned get_n_waypoints() const { return waypoints.size(); }
	const Waypoint &get_waypoint(unsigned) const;
	const TrainRouteMetric &get_metric(int = -1) const;
	void set_departure_delay(const Msp::Time::TimeDelta &);
	const Msp::Time::TimeDelta &get_departure_delay() const { return delay; }
	void set_trip_duration(const Msp::Time::TimeDelta &);
	const Msp::Time::TimeDelta &get_trip_duration() const { return duration; }

	virtual void message(const Message &);
	virtual void tick(const Msp::Time::TimeDelta &);

	void save(std::list<Msp::DataFile::Statement> &) const;

private:
	void block_reserved(Block &, Train *);
	void train_advanced(Block &);
	void train_rear_advanced(Block &);

	void create_metrics();
	bool create_lead_route();
	bool advance_to_track(RouteList::iterator &, const TrackIter &);

	static void get_routers(Layout &, std::vector<TrainRouter *> &, TrainRoutePlanner * = 0);
	static void start_planning(Layout &);
	static void apply_plan(Layout &, TrainRoutePlanner &);
};

} // namespace R2C2

#endif
